/* Pango
 * m17n-fc.c: Generic shaper using the m17n library for FreeType-based backends

   Copyright (C) 2004, 2005
     National Institute of Advanced Industrial Science and Technology (AIST)
     Registration Number H15PRO112

   This file is part of the m17n library.

   The m17n library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   The m17n library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the m17n library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.  */

#include <config.h>

#include <string.h>
#include <stdlib.h>

#include <pango/pango-engine.h>
#include <pango/pango-utils.h>
#include <pango/pangofc-font.h>

#include <m17n-gui.h>
#include <m17n-misc.h>

#ifdef G_LOG_DOMAIN
#undef G_LOG_DOMAIN
#endif
#define G_LOG_DOMAIN "M17N-Pango-FC"

/* Control printing of debug information.  It is initialized from the
   environment variable PANGO_M17N_FC_DEBUG.  Currently, any positive
   value turns on all debug messages.  */
static int debug_level;

/* Commonly used frame on nulldevice.  */
static MFrame *frame;

/* Initialize this module.  The main task is to open a frame on
   nulldevice.  */

static void
init_module ()
{
  const char *envvar = g_getenv ("PANGO_M17N_FC_DEBUG");
  MPlist *param;
  MFace *face;
  MFontset *fontset;

  debug_level = envvar ? atoi (envvar) : 0;

  param = mplist ();
  face = mface ();
  fontset = (mdatabase_find (Mfontset, msymbol ("generic"), Mnil, Mnil)
	     ? mfontset ("generic") : mfontset (NULL));
  mface_put_prop (face, Mfontset, fontset);
  mplist_put (param, Mface, face);
  mplist_put (param, Mdevice, Mnil);
  frame = mframe (param);
  m17n_object_unref (fontset);
  m17n_object_unref (face);
  m17n_object_unref (param);
}

/* Finalize this module by freeing frame.  */

static void
fini_module ()
{
  m17n_object_unref (frame);
}

/* No extra fields needed for these structures.  */
typedef PangoEngineShape      M17NEngineFc;
typedef PangoEngineShapeClass M17NEngineFcClass;

#define SCRIPT_ENGINE_NAME "M17NScriptEngineFc"
#define RENDER_TYPE PANGO_RENDER_TYPE_FC

static PangoEngineScriptInfo m17n_scripts[] = {
  { PANGO_SCRIPT_ARABIC,   "*" },
  { PANGO_SCRIPT_BENGALI,  "*" },
  { PANGO_SCRIPT_DEVANAGARI, "*" },
  { PANGO_SCRIPT_ETHIOPIC, "*" },
  { PANGO_SCRIPT_GEORGIAN, "*" },
  { PANGO_SCRIPT_GUJARATI, "*" },
  { PANGO_SCRIPT_GURMUKHI, "*" },
  { PANGO_SCRIPT_HEBREW,   "*" },
  { PANGO_SCRIPT_KANNADA,  "*" },
  { PANGO_SCRIPT_KHMER,    "*" },
  { PANGO_SCRIPT_LAO,      "*" },
  { PANGO_SCRIPT_MALAYALAM, "*" },
  { PANGO_SCRIPT_MYANMAR,  "*" },
  { PANGO_SCRIPT_ORIYA,     "*" },
  { PANGO_SCRIPT_SINHALA,   "*" },
  { PANGO_SCRIPT_SYRIAC,    "*" },
  { PANGO_SCRIPT_TAMIL,     "*" },
  { PANGO_SCRIPT_TELUGU,    "*" },
  { PANGO_SCRIPT_THAANA,    "*" },
  { PANGO_SCRIPT_THAI,      "*" },
  { PANGO_SCRIPT_TIBETAN,   "*" },
};

static PangoEngineInfo script_engines[] = {
  {
    SCRIPT_ENGINE_NAME,
    PANGO_ENGINE_TYPE_SHAPE,
    RENDER_TYPE,
    m17n_scripts, G_N_ELEMENTS (m17n_scripts)
  }
};


/* Convert PANGO_LANGUAGE to m17n-lib's language symbol.  If m17n-lib
   doesn't know about PANGO_LANGUAGE, return Mnil.  */

static MSymbol
m17n_fc_get_language (PangoLanguage *pango_language)
{
  MSymbol language;

  if (pango_language)
    {
      const char *lang = pango_language_to_string (pango_language);

      language = msymbol (lang);
      if (! msymbol_get (language, Mlanguage))
	{
	  if (strlen (lang) <= 2)
	    language = Mnil;
	  else
	    {
	      /* Remove region part (e.g. "zh_CN" -> "zh").  */
	      char shortlang[3];
	  
	      shortlang[0] = lang[0], shortlang[1] = lang[1], shortlang[2] = 0;
	      language = msymbol (shortlang);
	      if (! msymbol_get (language, Mlanguage))
		language = Mnil;
	    }
	}
    }
  else
    language = Mnil;
  return language;
}

/* Shaper function.  */

static void 
m17n_fc_engine_shape (PangoEngineShape *engine,
		      PangoFont        *pango_font,
		      const char       *text,
		      gint              length,
		      PangoAnalysis    *analysis,
		      PangoGlyphString *glyphs)
{
  /* Symbols for character property `category'.  */
  static MSymbol MZs, MCf;
  static MDrawControl control;
  MSymbol language;
  PangoGlyphInfo *g;
  MText *mt;
  MDrawGlyph *m_glyphs;
  int nchars, nglyphs;
  PangoFcFont *fc_font;
  char *fontname;
  MFont *font;
  int i;
  const char *p;
  int *offsets;

  g_return_if_fail (pango_font != NULL);
  g_return_if_fail (text != NULL);
  g_return_if_fail (length >= 0);
  g_return_if_fail (analysis != NULL);

  if (debug_level > 0)
    {
      char *msg = alloca (length + 1);

      if (msg)
	{
	  memcpy (msg, text, length);
	  msg[length] = '\0';
	}
      else
	msg = "...";
      g_debug ("shape \"%s\"", msg);
    }

  if (! MZs)
    {
      MZs = msymbol ("Zs");
      MCf = msymbol ("Cf");
    }

  language = m17n_fc_get_language (analysis->language);

  mt = mtext_from_data (text, length, MTEXT_FORMAT_UTF_8);
  nchars = mtext_len (mt);
  offsets = alloca (sizeof (int) * nchars);
  g_return_if_fail (offsets != NULL);
  for (i = 0, p = text; i < nchars; i++, p = g_utf8_next_char (p))
    offsets[i] = p - text;

  fc_font = PANGO_FC_FONT (pango_font);
  fontname = (char *) FcNameUnparse (fc_font->font_pattern);
  font = mfont_parse_name (fontname, Mfontconfig);
  free (fontname);
  {
    FcChar8 *filename;
    int size;

    if (FcPatternGetString (fc_font->font_pattern, FC_FILE, 0, &filename)
	== FcResultMatch)
      mfont_put_prop (font, Mfontfile, msymbol ((char *) filename));
    size = (int) mfont_get_prop (font, Msize);
    if (size)
      mfont_put_prop (font, Msize, (void *) (size * 256));
    if (debug_level > 0)
      g_debug ("  by %s", (char *) filename);
  }

  mtext_put_prop (mt, 0, nchars, Mfont, font);
  if (language != Mnil)
    mtext_put_prop (mt, 0, nchars, Mlanguage, language);

  control.enable_bidi = 1;
  m_glyphs = alloca (sizeof (MDrawGlyph) * nchars * 2);
  g_return_if_fail (m_glyphs != NULL);
  if (mdraw_glyph_list (frame, mt, 0, nchars, &control, m_glyphs,
			nchars * 2, &nglyphs) < 0)
    {
      m_glyphs = alloca (sizeof (MDrawGlyph) * nglyphs);
      g_return_if_fail (m_glyphs != NULL);
      mdraw_glyph_list (frame, mt, 0, nchars, &control, m_glyphs,
			nglyphs, &nglyphs);
    }

  pango_glyph_string_set_size (glyphs, nglyphs);

  for (i = 0, g = glyphs->glyphs; i < nglyphs; i++, g++)
    {
      if (m_glyphs[i].glyph_code >= 0)
	{
	  g->glyph = m_glyphs[i].glyph_code;
	  g->geometry.x_offset = m_glyphs[i].x_off * PANGO_SCALE / 256;
	  g->geometry.y_offset = m_glyphs[i].y_off * PANGO_SCALE / 256;
	  g->geometry.width = m_glyphs[i].x_advance * PANGO_SCALE / 256;
	}
      else
	{
	  int c = mtext_ref_char (mt, m_glyphs[i].from);
	  MSymbol category = mchar_get_prop (c, Mcategory);
	  PangoRectangle logical_rect;

	  if (category == MZs)
	    {
	      g->glyph = 0;
	      g->geometry.width = m_glyphs[i].x_advance * PANGO_SCALE / 256;
	    }
	  else if (category == MCf)
	    {
	      g->glyph = 0;
	      g->geometry.width = 0;
	    }
	  else
	    {
	      g->glyph = pango_fc_font_get_unknown_glyph (fc_font, c);
	      pango_font_get_glyph_extents (pango_font, g->glyph, NULL,
					    &logical_rect);
	      g->geometry.width = logical_rect.width;
	    }
	  g->geometry.x_offset = 0;
	  g->geometry.y_offset = 0;
	}
      g->attr.is_cluster_start
	= (i == 0
	   || m_glyphs[i - 1].from != m_glyphs[i].from);
      glyphs->log_clusters[i] = offsets[m_glyphs[i].from];
    }

  m17n_object_unref (mt);
  free (font);
}

/* Check if PANGO_FONT is suitable for the character WC.  */

static PangoCoverageLevel
m17n_fc_engine_covers (PangoEngineShape *engine,
		       PangoFont        *pango_font,
		       PangoLanguage    *pango_language,
		       gunichar          wc)
{
  static gunichar last_wc = 0;
  /* Symbols for character property `script'.  */
  static MSymbol Mlatin, Mcommon, Minherited;
  PangoCoverage *coverage;
  PangoCoverageLevel result;
  PangoFcFont *fc_font;
  char *fontname;
  MFont *font;
  MSymbol script, language;
  int check_result;

  if (! Mlatin)
    {
      Mlatin = msymbol ("latin");
      Mcommon = msymbol ("common");
      Minherited = msymbol ("inherited");
    }

  if (debug_level > 0 && wc != last_wc)
    {
      g_debug ("covers for U+%04X", wc);
      last_wc = wc;
    }

  coverage = pango_font_get_coverage (pango_font, pango_language);
  result = pango_coverage_get (coverage, wc);
  pango_coverage_unref (coverage);

  if (result == PANGO_COVERAGE_NONE)
    {
      if (debug_level > 0)
	{
	  fc_font = PANGO_FC_FONT (pango_font);
	  fontname = (char *) FcNameUnparse (fc_font->font_pattern);
	  font = mfont_parse_name (fontname, Mfontconfig);
	  g_debug ("  %s none", 
		   msymbol_name (mfont_get_prop (font, Mfamily)));
	  free (fontname);
	  free (font);
	}
      return result;
    }

  fc_font = PANGO_FC_FONT (pango_font);
  fontname = (char *) FcNameUnparse (fc_font->font_pattern);
  font = mfont_parse_name (fontname, Mfontconfig);

  if (wc < 0x100)
    script = Mlatin;
  else
    {
      script = mchar_get_prop ((int) wc, Mscript);
      if (script == Mcommon || script == Minherited)
	script = mchar_get_prop ((int) wc, msymbol ("block"));
    }
  language = m17n_fc_get_language (pango_language);

  check_result = mfont_check (frame, NULL, script, language, font);
  if (debug_level > 0)
    {
      char *result_text = (check_result == 2 ? "exact"
			   : check_result == 1 ? "approximate"
			   : "fallback");

      g_debug ("  (%s,%s): %s %s",
	       (script ? msymbol_name (script) : ""),
	       (language ? msymbol_name (language) : ""),
	       msymbol_name (mfont_get_prop (font, Mfamily)),
	       result_text);
    }      

  free (fontname);
  free (font);
  return (check_result == 2 ? PANGO_COVERAGE_EXACT
	  : check_result == 1 ? PANGO_COVERAGE_APPROXIMATE
	  : PANGO_COVERAGE_FALLBACK);
}

static void
m17n_engine_fc_class_init (PangoEngineShapeClass *class)
{
  class->script_shape = m17n_fc_engine_shape;
  class->covers = m17n_fc_engine_covers;
}

PANGO_ENGINE_SHAPE_DEFINE_TYPE (M17NEngineFc, m17n_engine_fc,
				m17n_engine_fc_class_init, NULL)

void 
PANGO_MODULE_ENTRY(init) (GTypeModule *module)
{
  m17n_engine_fc_register_type (module);

  M17N_INIT ();
  init_module ();
}

void 
PANGO_MODULE_ENTRY(exit) (void)
{
  fini_module ();
  M17N_FINI ();
}

void 
PANGO_MODULE_ENTRY(list) (PangoEngineInfo **engines,
			  int              *n_engines)
{
  *engines = script_engines;
  *n_engines = G_N_ELEMENTS (script_engines);
}

PangoEngine *
PANGO_MODULE_ENTRY(create) (const char *id)
{
  return (!strcmp (id, SCRIPT_ENGINE_NAME)
	  ? g_object_new (m17n_engine_fc_type, NULL)
	  : NULL);
}
